@@ -41,9 +41,9 @@ var (
 	SnapDataHomeGlob          string
 	SnapDownloadCacheDir      string
 	SnapAppArmorDir           string
-	AppArmorCacheDir          string
 	SnapAppArmorAdditionalDir string
 	SnapConfineAppArmorDir    string
+	SnapSeccompBase           string
 	SnapSeccompDir            string
 	SnapMountPolicyDir        string
 	SnapUdevRulesDir          string
@@ -55,6 +55,9 @@ var (
 	SnapRunDir                string
 	SnapRunNsDir              string
 	SnapRunLockDir            string
+	SnapBootstrapRunDir       string
+
+	SnapdStoreSSLCertsDir string
 
 	SnapSeedDir   string
 	SnapDeviceDir string
@@ -74,19 +77,23 @@ var (
 	SnapRepairAssertsDir string
 	SnapRunRepairDir     string
 
-	SnapCacheDir     string
-	SnapNamesFile    string
-	SnapSectionsFile string
-	SnapCommandsDB   string
+	SnapRollbackDir string
+
+	SnapCacheDir        string
+	SnapNamesFile       string
+	SnapSectionsFile    string
+	SnapCommandsDB      string
+	SnapAuxStoreInfoDir string
 
 	SnapBinariesDir     string
 	SnapServicesDir     string
+	SnapUserServicesDir string
 	SnapSystemdConfDir  string
 	SnapDesktopFilesDir string
+	SnapDesktopIconsDir string
 	SnapBusPolicyDir    string
 
-	SystemApparmorDir      string
-	SystemApparmorCacheDir string
+	SnapModeenvFile string
 
 	CloudMetaDataFile     string
 	CloudInstanceDataFile string
@@ -96,15 +103,17 @@ var (
 	XdgRuntimeDirBase string
 	XdgRuntimeDirGlob string
 
-	CompletionHelper string
-	CompletersDir    string
+	CompletionHelperInCore string
+	CompletersDir          string
 
-	SystemFontsDir           string
-	SystemLocalFontsDir      string
-	SystemFontconfigCacheDir string
+	SystemFontsDir            string
+	SystemLocalFontsDir       string
+	SystemFontconfigCacheDirs []string
 
 	FreezerCgroupDir string
-	SnapshotsDir     string
+	PidsCgroupDir    string
+
+	SnapshotsDir string
 
 	ErrtrackerDbDir string
 	SysfsDir        string
@@ -133,6 +142,8 @@ const (
 var (
 	// not exported because it does not honor the global rootdir
 	snappyDir = filepath.Join("var", "lib", "snapd")
+
+	callbacks = []func(string){}
 )
 
 func init() {
@@ -159,14 +170,19 @@ func StripRootDir(dir string) string {
 
 // SupportsClassicConfinement returns true if the current directory layout supports classic confinement.
 func SupportsClassicConfinement() bool {
+	// Core systems don't support classic confinement as a policy decision.
+	if !release.OnClassic {
+		return false
+	}
+
+	// Classic systems support classic confinement if using the primary mount
+	// location for snaps, that is /snap or if using the alternate mount
+	// location, /var/lib/snapd/snap along with the /snap ->
+	// /var/lib/snapd/snap symlink in place.
 	smd := filepath.Join(GlobalRootDir, defaultSnapMountDir)
 	if SnapMountDir == smd {
 		return true
 	}
-
-	// distros with a non-default /snap location may still be good
-	// if there is a symlink in place that links from the
-	// defaultSnapMountDir (/snap) to the distro specific mount dir
 	fi, err := os.Lstat(smd)
 	if err == nil && fi.Mode()&os.ModeSymlink != 0 {
 		if target, err := filepath.EvalSymlinks(smd); err == nil {
@@ -194,6 +210,44 @@ func isInsideBaseSnap() (bool, error) {
 	return err == nil, err
 }
 
+// SnapBlobDirUnder returns the path to the snap blob dir under rootdir.
+func SnapBlobDirUnder(rootdir string) string {
+	return filepath.Join(rootdir, snappyDir, "snaps")
+}
+
+// SnapSeedDirUnder returns the path to the snap seed dir under rootdir.
+func SnapSeedDirUnder(rootdir string) string {
+	return filepath.Join(rootdir, snappyDir, "seed")
+}
+
+// SnapStateFileUnder returns the path to snapd state file under rootdir.
+func SnapStateFileUnder(rootdir string) string {
+	return filepath.Join(rootdir, snappyDir, "state.json")
+}
+
+// SnapModeenvFileUnder returns the path to the modeenv file under rootdir.
+func SnapModeenvFileUnder(rootdir string) string {
+	return filepath.Join(rootdir, snappyDir, "modeenv")
+}
+
+// FeaturesDirUnder returns the path to the features dir under rootdir.
+func FeaturesDirUnder(rootdir string) string {
+	return filepath.Join(rootdir, snappyDir, "features")
+}
+
+// SnapSystemdConfDirUnder returns the path to the systemd conf dir under
+// rootdir.
+func SnapSystemdConfDirUnder(rootdir string) string {
+	return filepath.Join(rootdir, "/etc/systemd/system.conf.d")
+}
+
+// AddRootDirCallback registers a callback for whenever the global root
+// directory (set by SetRootDir) is changed to enable updates to variables in
+// other packages that depend on its location.
+func AddRootDirCallback(c func(string)) {
+	callbacks = append(callbacks, c)
+}
+
 // SetRootDir allows settings a new global root directory, this is useful
 // for e.g. chroot operations
 func SetRootDir(rootdir string) {
@@ -202,8 +256,17 @@ func SetRootDir(rootdir string) {
 	}
 	GlobalRootDir = rootdir
 
+	altDirDistros := []string{
+		"antergos",
+		"arch",
+		"archlinux",
+		"fedora",
+		"manjaro",
+		"manjaro-arm",
+	}
+
 	isInsideBase, _ := isInsideBaseSnap()
-	if !isInsideBase && release.DistroLike("fedora", "arch", "archlinux", "manjaro", "antergos") {
+	if !isInsideBase && release.DistroLike(altDirDistros...) {
 		SnapMountDir = filepath.Join(rootdir, "/var/lib/snapd/snap")
 	} else {
 		SnapMountDir = filepath.Join(rootdir, defaultSnapMountDir)
@@ -213,18 +276,26 @@ func SetRootDir(rootdir string) {
 	SnapDataHomeGlob = filepath.Join(rootdir, "/home/*/", UserHomeSnapDir)
 	SnapAppArmorDir = filepath.Join(rootdir, snappyDir, "apparmor", "profiles")
 	SnapConfineAppArmorDir = filepath.Join(rootdir, snappyDir, "apparmor", "snap-confine")
-	AppArmorCacheDir = filepath.Join(rootdir, "/var/cache/apparmor")
 	SnapAppArmorAdditionalDir = filepath.Join(rootdir, snappyDir, "apparmor", "additional")
 	SnapDownloadCacheDir = filepath.Join(rootdir, snappyDir, "cache")
-	SnapSeccompDir = filepath.Join(rootdir, snappyDir, "seccomp", "bpf")
+	SnapSeccompBase = filepath.Join(rootdir, snappyDir, "seccomp")
+	SnapSeccompDir = filepath.Join(SnapSeccompBase, "bpf")
 	SnapMountPolicyDir = filepath.Join(rootdir, snappyDir, "mount")
 	SnapMetaDir = filepath.Join(rootdir, snappyDir, "meta")
-	SnapBlobDir = filepath.Join(rootdir, snappyDir, "snaps")
+	SnapBlobDir = SnapBlobDirUnder(rootdir)
+	// ${snappyDir}/desktop is added to $XDG_DATA_DIRS.
+	// Subdirectories are interpreted according to the relevant
+	// freedesktop.org specifications
 	SnapDesktopFilesDir = filepath.Join(rootdir, snappyDir, "desktop", "applications")
+	SnapDesktopIconsDir = filepath.Join(rootdir, snappyDir, "desktop", "icons")
 	SnapRunDir = filepath.Join(rootdir, "/run/snapd")
 	SnapRunNsDir = filepath.Join(SnapRunDir, "/ns")
 	SnapRunLockDir = filepath.Join(SnapRunDir, "/lock")
 
+	SnapBootstrapRunDir = filepath.Join(SnapRunDir, "snap-bootstrap")
+
+	SnapdStoreSSLCertsDir = filepath.Join(rootdir, snappyDir, "ssl/store-certs")
+
 	// keep in sync with the debian/snapd.socket file:
 	SnapdSocket = filepath.Join(rootdir, "/run/snapd.socket")
 	SnapSocket = filepath.Join(rootdir, "/run/snapd-snap.socket")
@@ -234,31 +305,34 @@ func SetRootDir(rootdir string) {
 	SnapAssertsSpoolDir = filepath.Join(rootdir, "run/snapd/auto-import")
 	SnapSeqDir = filepath.Join(rootdir, snappyDir, "sequence")
 
-	SnapStateFile = filepath.Join(rootdir, snappyDir, "state.json")
+	SnapStateFile = SnapStateFileUnder(rootdir)
 	SnapSystemKeyFile = filepath.Join(rootdir, snappyDir, "system-key")
 
 	SnapCacheDir = filepath.Join(rootdir, "/var/cache/snapd")
 	SnapNamesFile = filepath.Join(SnapCacheDir, "names")
 	SnapSectionsFile = filepath.Join(SnapCacheDir, "sections")
 	SnapCommandsDB = filepath.Join(SnapCacheDir, "commands.db")
+	SnapAuxStoreInfoDir = filepath.Join(SnapCacheDir, "aux")
 
-	SnapSeedDir = filepath.Join(rootdir, snappyDir, "seed")
+	SnapSeedDir = SnapSeedDirUnder(rootdir)
 	SnapDeviceDir = filepath.Join(rootdir, snappyDir, "device")
 
+	SnapModeenvFile = SnapModeenvFileUnder(rootdir)
+
 	SnapRepairDir = filepath.Join(rootdir, snappyDir, "repair")
 	SnapRepairStateFile = filepath.Join(SnapRepairDir, "repair.json")
 	SnapRepairRunDir = filepath.Join(SnapRepairDir, "run")
 	SnapRepairAssertsDir = filepath.Join(SnapRepairDir, "assertions")
 	SnapRunRepairDir = filepath.Join(SnapRunDir, "repair")
 
+	SnapRollbackDir = filepath.Join(rootdir, snappyDir, "rollback")
+
 	SnapBinariesDir = filepath.Join(SnapMountDir, "bin")
 	SnapServicesDir = filepath.Join(rootdir, "/etc/systemd/system")
-	SnapSystemdConfDir = filepath.Join(rootdir, "/etc/systemd/system.conf.d")
+	SnapUserServicesDir = filepath.Join(rootdir, "/etc/systemd/user")
+	SnapSystemdConfDir = SnapSystemdConfDirUnder(rootdir)
 	SnapBusPolicyDir = filepath.Join(rootdir, "/etc/dbus-1/system.d")
 
-	SystemApparmorDir = filepath.Join(rootdir, "/etc/apparmor.d")
-	SystemApparmorCacheDir = filepath.Join(rootdir, "/etc/apparmor.d/cache")
-
 	CloudMetaDataFile = filepath.Join(rootdir, "/var/lib/cloud/seed/nocloud-net/meta-data")
 	CloudInstanceDataFile = filepath.Join(rootdir, "/run/cloud-init/instance-data.json")
 
@@ -278,16 +352,16 @@ func SetRootDir(rootdir string) {
 	}
 
 	XdgRuntimeDirBase = filepath.Join(rootdir, "/run/user")
-	XdgRuntimeDirGlob = filepath.Join(rootdir, XdgRuntimeDirBase, "*/")
+	XdgRuntimeDirGlob = filepath.Join(XdgRuntimeDirBase, "*/")
 
-	CompletionHelper = filepath.Join(CoreLibExecDir, "etelpmoc.sh")
+	CompletionHelperInCore = filepath.Join(CoreLibExecDir, "etelpmoc.sh")
 	CompletersDir = filepath.Join(rootdir, "/usr/share/bash-completion/completions/")
 
 	// These paths agree across all supported distros
 	SystemFontsDir = filepath.Join(rootdir, "/usr/share/fonts")
 	SystemLocalFontsDir = filepath.Join(rootdir, "/usr/local/share/fonts")
 	// The cache path is true for Ubuntu, Debian, openSUSE, Arch
-	SystemFontconfigCacheDir = filepath.Join(rootdir, "/var/cache/fontconfig")
+	SystemFontconfigCacheDirs = []string{filepath.Join(rootdir, "/var/cache/fontconfig")}
 	if release.DistroLike("fedora") && !release.DistroLike("amzn") {
 		// Applies to Fedora and CentOS, Amazon Linux 2 is behind with
 		// updates to fontconfig and uses /var/cache/fontconfig instead,
@@ -295,16 +369,28 @@ func SetRootDir(rootdir string) {
 		// https://fedoraproject.org/wiki/Changes/FontconfigCacheDirChange
 		// https://bugzilla.redhat.com/show_bug.cgi?id=1416380
 		// https://bugzilla.redhat.com/show_bug.cgi?id=1377367
-		SystemFontconfigCacheDir = filepath.Join(rootdir, "/usr/lib/fontconfig/cache")
+		//
+		// However, snaps may still use older libfontconfig, which fails
+		// to parse the new config and defaults to
+		// /var/cache/fontconfig. In this case we need to make both
+		// locations available
+		SystemFontconfigCacheDirs = append(SystemFontconfigCacheDirs, filepath.Join(rootdir, "/usr/lib/fontconfig/cache"))
 	}
 
 	FreezerCgroupDir = filepath.Join(rootdir, "/sys/fs/cgroup/freezer/")
+	PidsCgroupDir = filepath.Join(rootdir, "/sys/fs/cgroup/pids/")
 	SnapshotsDir = filepath.Join(rootdir, snappyDir, "snapshots")
 
 	ErrtrackerDbDir = filepath.Join(rootdir, snappyDir, "errtracker.db")
 	SysfsDir = filepath.Join(rootdir, "/sys")
 
-	FeaturesDir = filepath.Join(rootdir, snappyDir, "features")
+	FeaturesDir = FeaturesDirUnder(rootdir)
+
+	// call the callbacks last so that the callbacks can just reference the
+	// global vars if they want, instead of using the new rootdir directly
+	for _, c := range callbacks {
+		c(rootdir)
+	}
 }
 
 // what inside a (non-classic) snap is /usr/lib/snapd, outside can come from different places
@@ -328,5 +414,6 @@ func CompleteShPath(base string) string {
 
 func IsCompleteShSymlink(compPath string) bool {
 	target, err := os.Readlink(compPath)
-	return err == nil && filepath.Base(target) == "complete.sh"
+	// check if the target paths ends with "/snapd/complete.sh"
+	return err == nil && filepath.Base(filepath.Dir(target)) == "snapd" && filepath.Base(target) == "complete.sh"
 }
