@@ -54,6 +54,9 @@ amazon_name_package() {
             xdelta3)
                 echo "xdelta"
                 ;;
+            openvswitch-switch)
+                echo "openvswitch"
+                ;;
             *)
                 echo "$i"
                 ;;
@@ -125,12 +128,12 @@ distro_name_package() {
         ubuntu-*|debian-*)
             debian_name_package "$@"
             ;;
-        fedora-*)
-            fedora_name_package "$@"
-            ;;
-        amazon-*|centos-*)
+        amazon-*|centos-7-*)
             amazon_name_package "$@"
             ;;
+        fedora-*|centos-*)
+            fedora_name_package "$@"
+            ;;
         opensuse-*)
             opensuse_name_package "$@"
             ;;
@@ -160,25 +163,25 @@ distro_install_local_package() {
     case "$SPREAD_SYSTEM" in
         ubuntu-14.04-*|debian-*)
             # relying on dpkg as apt(-get) does not support installation from local files in trusty.
-            dpkg -i --force-depends --auto-deconfigure --force-depends-version "$@"
-            apt-get -f install -y
+            eatmydata dpkg -i --force-depends --auto-deconfigure --force-depends-version "$@"
+            eatmydata apt-get -f install -y
             ;;
         ubuntu-*)
-            flags="-y"
+            flags="-y --no-install-recommends"
             if [ "$allow_downgrades" = "true" ]; then
                 flags="$flags --allow-downgrades"
             fi
             # shellcheck disable=SC2086
             apt install $flags "$@"
             ;;
-        fedora-*)
-            quiet dnf -y install --setopt=install_weak_deps=False "$@"
-            ;;
-        amazon-*|centos-*)
+        amazon-*|centos-7-*)
             quiet yum -y localinstall "$@"
             ;;
+        fedora-*|centos-*)
+            quiet dnf -y install --setopt=install_weak_deps=False "$@"
+            ;;
         opensuse-*)
-            quiet rpm -i "$@"
+            quiet rpm -i --replacepkgs "$@"
             ;;
         arch-*)
             pacman -U --noconfirm "$@"
@@ -233,17 +236,17 @@ distro_install_package() {
     case "$SPREAD_SYSTEM" in
         ubuntu-*|debian-*)
         if [[ "$*" =~ "libudev-dev" ]]; then
-            apt-get install -y --only-upgrade systemd
+            eatmydata apt-get install -y --only-upgrade systemd
         fi
         ;;
     esac
 
-    # fix dependency issue where libp11-kit0 needs to be downgraded to 
+    # fix dependency issue where libp11-kit0 needs to be downgraded to
     # install gnome-keyring
     case "$SPREAD_SYSTEM" in
         debian-9-*)
         if [[ "$*" =~ "gnome-keyring" ]]; then
-            apt-get remove -y libp11-kit0
+            eatmydata apt-get remove -y libp11-kit0
         fi
         ;;
     esac
@@ -264,19 +267,27 @@ distro_install_package() {
     case "$SPREAD_SYSTEM" in
         ubuntu-*|debian-*)
             # shellcheck disable=SC2086
-            quiet apt-get install $APT_FLAGS -y "${pkg_names[@]}"
+            quiet eatmydata apt-get install $APT_FLAGS -y "${pkg_names[@]}"
             ;;
-        fedora-*)
-            # shellcheck disable=SC2086
-            quiet dnf -y --refresh install $DNF_FLAGS "${pkg_names[@]}"
-            ;;
-        amazon-*|centos-*)
+        amazon-*|centos-7-*)
             # shellcheck disable=SC2086
             quiet yum -y install $YUM_FLAGS "${pkg_names[@]}"
             ;;
-        opensuse-*)
+        fedora-*|centos-*)
             # shellcheck disable=SC2086
-            quiet zypper install -y $ZYPPER_FLAGS "${pkg_names[@]}"
+            quiet dnf -y --refresh install $DNF_FLAGS "${pkg_names[@]}"
+            ;;
+        opensuse-*)
+            # packages may be downgraded in the repositories, which would be
+            # picked up next time we ran `zypper dup` and applied locally;
+            # however we only update the images periodically, in the meantime,
+            # when downgrades affect packages we need or have installed, `zypper
+            # in` may stop with the prompt asking the user about either breaking
+            # the installed packages or allowing downgrades, passing
+            # --allow-downgrade will make the installation proceed
+
+            # shellcheck disable=SC2086
+            quiet zypper install -y --allow-downgrade --force-resolution $ZYPPER_FLAGS "${pkg_names[@]}"
             ;;
         arch-*)
             # shellcheck disable=SC2086
@@ -306,15 +317,15 @@ distro_purge_package() {
 
     case "$SPREAD_SYSTEM" in
         ubuntu-*|debian-*)
-            quiet apt-get remove -y --purge -y "$@"
+            quiet eatmydata apt-get remove -y --purge -y "$@"
             ;;
-        fedora-*)
+        amazon-*|centos-7-*)
+            quiet yum -y remove "$@"
+            ;;
+        fedora-*|centos-*)
             quiet dnf -y remove "$@"
             quiet dnf clean all
             ;;
-        amazon-*|centos-*)
-            quiet yum -y remove "$@"
-            ;;
         opensuse-*)
             quiet zypper remove -y "$@"
             ;;
@@ -331,16 +342,16 @@ distro_purge_package() {
 distro_update_package_db() {
     case "$SPREAD_SYSTEM" in
         ubuntu-*|debian-*)
-            quiet apt-get update
+            quiet eatmydata apt-get update
             ;;
-        fedora-*)
-            quiet dnf clean all
-            quiet dnf makecache
-            ;;
-        amazon-*|centos-*)
+        amazon-*|centos-7-*)
             quiet yum clean all
             quiet yum makecache
             ;;
+        fedora-*|centos-*)
+            quiet dnf clean all
+            quiet dnf makecache
+            ;;
         opensuse-*)
             quiet zypper --gpg-auto-import-keys refresh
             ;;
@@ -357,14 +368,14 @@ distro_update_package_db() {
 distro_clean_package_cache() {
     case "$SPREAD_SYSTEM" in
         ubuntu-*|debian-*)
-            quiet apt-get clean
+            quiet eatmydata apt-get clean
             ;;
-        fedora-*)
-            dnf clean all
-            ;;
-        amazon-*|centos-*)
+        amazon-*|centos-7-*)
             yum clean all
             ;;
+        fedora-*|centos-*)
+            dnf clean all
+            ;;
         opensuse-*)
             zypper -q clean --all
             ;;
@@ -381,14 +392,14 @@ distro_clean_package_cache() {
 distro_auto_remove_packages() {
     case "$SPREAD_SYSTEM" in
         ubuntu-*|debian-*)
-            quiet apt-get -y autoremove
+            quiet eatmydata apt-get -y autoremove
             ;;
-        fedora-*)
-            quiet dnf -y autoremove
-            ;;
-        amazon-*|centos-*)
+        amazon-*|centos-7-*)
             quiet yum -y autoremove
             ;;
+        fedora-*|centos-*)
+            quiet dnf -y autoremove
+            ;;
         opensuse-*)
             ;;
         arch-*)
@@ -405,12 +416,12 @@ distro_query_package_info() {
         ubuntu-*|debian-*)
             apt-cache policy "$1"
             ;;
-        fedora-*)
-            dnf info "$1"
-            ;;
-        amazon-*|centos-*)
+        amazon-*|centos-7-*)
             yum info "$1"
             ;;
+        fedora-*|centos-*)
+            dnf info "$1"
+            ;;
         opensuse-*)
             zypper info "$1"
             ;;
@@ -433,11 +444,22 @@ distro_install_build_snapd(){
         apt install -y --only-upgrade snapd
         mv sources.list.back /etc/apt/sources.list
         apt update
+
         # On trusty we may pull in a new hwe-kernel that is needed to run the
         # snapd tests. We need to reboot to actually run this kernel.
         if [[ "$SPREAD_SYSTEM" = ubuntu-14.04-* ]] && [ "$SPREAD_REBOOT" = 0 ]; then
             REBOOT
         fi
+    elif [ -n "$PPA_VALIDATION_NAME" ]; then
+        apt install -y snapd
+        add-apt-repository -y "$PPA_VALIDATION_NAME"
+        apt update
+        apt install -y --only-upgrade snapd
+        add-apt-repository --remove "$PPA_VALIDATION_NAME"
+        apt update
+
+        # Double check that it really comes from the PPA
+        apt show snapd | grep "APT-Sources: http.*ppa.launchpad.net"
     else
         packages=
         case "$SPREAD_SYSTEM" in
@@ -465,6 +487,26 @@ distro_install_build_snapd(){
         # shellcheck disable=SC2086
         distro_install_local_package $packages
 
+        case "$SPREAD_SYSTEM" in
+            fedora-*|centos-*)
+                # We need to wait until the man db cache is updated before do daemon-reexec
+                # Otherwise the service fails and the system will be degraded during tests executions
+                for i in $(seq 20); do
+                    if ! systemctl is-active run-*.service; then
+                        break
+                    fi
+                    sleep .5
+                done
+
+                # systemd caches SELinux policy data and subsequently attempts
+                # to create sockets with incorrect context, this installation of
+                # socket activated snaps fails, see:
+                # https://bugzilla.redhat.com/show_bug.cgi?id=1660141
+                # https://github.com/systemd/systemd/issues/9997
+                systemctl daemon-reexec
+                ;;
+        esac
+
         if [[ "$SPREAD_SYSTEM" == arch-* ]]; then
             # Arch policy does not allow calling daemon-reloads in package
             # install scripts
@@ -538,16 +580,20 @@ pkg_dependencies_ubuntu_classic(){
         avahi-daemon
         cups
         dbus-x11
+        fontconfig
         gnome-keyring
         jq
         man
         nfs-kernel-server
         printer-driver-cups-pdf
+        python3-dbus
+        python3-gi
         python3-yaml
         upower
         weston
         xdg-user-dirs
         xdg-utils
+        zsh
         "
 
     case "$SPREAD_SYSTEM" in
@@ -558,35 +604,49 @@ pkg_dependencies_ubuntu_classic(){
             echo "
                 gccgo-6
                 evolution-data-server
+                fwupd
                 gnome-online-accounts
+                packagekit
                 "
                 pkg_linux_image_extra
             ;;
         ubuntu-16.04-64)
             echo "
                 evolution-data-server
+                fwupd
                 gccgo-6
                 gnome-online-accounts
                 kpartx
                 libvirt-bin
+                packagekit
                 qemu
                 x11-utils
                 xvfb
                 "
                 pkg_linux_image_extra
             ;;
-        ubuntu-17.10-64)
-                pkg_linux_image_extra
-            ;;
         ubuntu-18.04-64)
             echo "
                 gccgo-8
                 evolution-data-server
+                fwupd
+                packagekit
                 "
             ;;
-        ubuntu-18.10-64)
+        ubuntu-19.10-64)
             echo "
                 evolution-data-server
+                fwupd
+                packagekit
+                qemu-utils
+                "
+            ;;
+        ubuntu-20.04-64)
+            echo "
+                evolution-data-server
+                gccgo-9
+                packagekit
+                qemu-utils
                 "
             ;;
         ubuntu-*)
@@ -596,8 +656,13 @@ pkg_dependencies_ubuntu_classic(){
             ;;
         debian-*)
             echo "
+                eatmydata
                 evolution-data-server
+                fwupd
+                libc6-dev-i386
                 net-tools
+                packagekit
+                sbuild
                 "
             ;;
     esac
@@ -630,21 +695,28 @@ pkg_dependencies_fedora(){
         dbus-x11
         evolution-data-server
         expect
+        fontconfig
+        fwupd
         git
         golang
         jq
         iptables-services
         man
-        mock
         net-tools
+        nmap-ncat
         nfs-utils
+        PackageKit
         python3-yaml
+        python3-dbus
+        python3-gobject
         redhat-lsb-core
         rpm-build
         udisks2
+        upower
         xdg-user-dirs
         xdg-utils
         strace
+        zsh
         "
 }
 
@@ -653,21 +725,25 @@ pkg_dependencies_amazon(){
         curl
         dbus-x11
         expect
+        fontconfig
+        fwupd
         git
         golang
         grub2-tools
         jq
         iptables-services
         man
-        mock
         nc
         net-tools
         nfs-utils
+        PackageKit
         system-lsb-core
         rpm-build
         xdg-user-dirs
         xdg-utils
         udisks2
+        upower
+        zsh
         "
 }
 
@@ -678,19 +754,24 @@ pkg_dependencies_opensuse(){
         curl
         evolution-data-server
         expect
+        fontconfig
+        fwupd
         git
         golang-packaging
         jq
         lsb-release
         man
         nfs-kernel-server
+        PackageKit
         python3-yaml
         netcat-openbsd
         osc
         udisks2
+        upower
         uuidd
         xdg-user-dirs
         xdg-utils
+        zsh
         "
 }
 
@@ -702,6 +783,8 @@ pkg_dependencies_arch(){
     curl
     evolution-data-server
     expect
+    fontconfig
+    fwupd
     git
     go
     go-tools
@@ -709,20 +792,26 @@ pkg_dependencies_arch(){
     libseccomp
     libcap
     libx11
+    man
     net-tools
     nfs-utils
     openbsd-netcat
+    packagekit
     python
     python-docutils
+    python-dbus
+    python-gobject
     python3-yaml
     squashfs-tools
     shellcheck
     strace
     udisks2
+    upower
     xdg-user-dirs
     xdg-utils
     xfsprogs
     apparmor
+    zsh
     "
 }
 
@@ -736,12 +825,12 @@ pkg_dependencies(){
             pkg_dependencies_ubuntu_generic
             pkg_dependencies_ubuntu_classic
             ;;
-        fedora-*)
-            pkg_dependencies_fedora
-            ;;
-        amazon-*|centos-*)
+        amazon-*|centos-7-*)
             pkg_dependencies_amazon
             ;;
+        fedora-*|centos-*)
+            pkg_dependencies_fedora
+            ;;
         opensuse-*)
             pkg_dependencies_opensuse
             ;;
